@@ -34,6 +34,7 @@
 #include "mb/pg_wchar.h"
 #include "miscadmin.h"
 #include "nodes/makefuncs.h"
+#include "nodes/nodeFuncs.h"
 #include "nodes/pg_list.h"
 #include "optimizer/cost.h"
 #include "optimizer/pathnode.h"
@@ -128,6 +129,7 @@ struct OracleFdwState {
 	char *nls_lang;                /* Oracle locale information */
 	oracleSession *session;        /* encapsulates the active Oracle session */
 	char *query;                   /* query we issue against Oracle */
+	List *params;                  /* list of parameters needed for the query */
 	struct paramDesc *paramList;   /* description of parameters needed for the query */
 	struct oraTable *oraTable;     /* description of the remote Oracle table */
 	Cost startup_cost;             /* cost estimate, only needed for planning */
@@ -175,12 +177,12 @@ static void oracleEndForeignScan(ForeignScanState *node);
  */
 static struct OracleFdwState *getFdwState(Oid foreigntableid, bool *plan_costs);
 static void oracleGetOptions(Oid foreigntableid, List **options);
-static char *createQuery(oracleSession *session, List *columnlist, List *conditions, struct oraTable *oraTable, struct paramDesc **paramList, handleClause **handle_clauses);
+static char *createQuery(oracleSession *session, RelOptInfo *foreignrel, struct oraTable *oraTable, List **params, handleClause **handle_clauses);
 static void getColumnData(Oid foreigntableid, struct oraTable *oraTable);
 #ifndef OLD_FDW_API
 static int acquireSampleRowsFunc (Relation relation, int elevel, HeapTuple *rows, int targrows, double *totalrows, double *totaldeadrows);
 #endif
-static bool getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const struct oraTable *oraTable, struct paramDesc **paramList);
+static bool getOracleWhereClause(oracleSession *session, RelOptInfo *foreignrel, char **where, Expr *expr, const struct oraTable *oraTable, List **params);
 static char *datumToString(Datum datum, Oid type);
 static void getUsedColumns(Expr *expr, struct oraTable *oraTable);
 static void checkDataTypes(struct oraTable *oraTable);
@@ -193,7 +195,7 @@ static char *deserializeString(Const *constant);
 static long deserializeLong(Const *constant);
 static void cleanupTransaction(ResourceReleasePhase phase, bool isCommit, bool isTopLevel, void *arg);
 static void exitHook(int code, Datum arg);
-static char *setParameters(struct paramDesc *paramList, EState *execstate);
+static char *setParameters(struct paramDesc *paramList, EState *execstate, ExprContext *econtext);
 static void convertTuple(struct OracleFdwState *fdw_state, Datum *values, bool *nulls, bool trunc_lob);
 static void errorContextCallback(void *arg);
 
@@ -382,7 +384,7 @@ oraclePlanForeignScan(Oid foreigntableid,
 	fdwState = getFdwState(foreigntableid, &plan_costs);
 
 	/* construct Oracle query and get the list of parameters and actions for RestrictInfos */
-	fdwState->query = createQuery(fdwState->session, baserel->reltargetlist, baserel->baserestrictinfo, fdwState->oraTable, &(fdwState->paramList), &(fdwState->handle_clauses));
+	fdwState->query = createQuery(fdwState->session, baserel, fdwState->oraTable, &(fdwState->params), &(fdwState->handle_clauses));
 	elog(DEBUG1, "oracle_fdw: remote query is: %s", fdwState->query);
 
 	/* get PostgreSQL column data types, check that they match Oracle's */
@@ -431,7 +433,7 @@ oracleGetForeignRelSize(PlannerInfo *root, RelOptInfo *baserel, Oid foreigntable
 	fdwState = getFdwState(foreigntableid, &plan_costs);
 
 	/* construct Oracle query and get the list of parameters and antions for RestrictInfos */
-	fdwState->query = createQuery(fdwState->session, baserel->reltargetlist, baserel->baserestrictinfo, fdwState->oraTable, &(fdwState->paramList), &(fdwState->handle_clauses));
+	fdwState->query = createQuery(fdwState->session, baserel, fdwState->oraTable, &(fdwState->params), &(fdwState->handle_clauses));
 	elog(DEBUG1, "oracle_fdw: remote query is: %s", fdwState->query);
 
 	/* get PostgreSQL column data types, check that they match Oracle's */
@@ -496,22 +498,32 @@ ForeignScan
 {
 	struct OracleFdwState *fdwState = (struct OracleFdwState *)baserel->fdw_private;
 	List *fdw_private, *keep_clauses = NIL;
+	ListCell *cell1, *cell2;
 	int i;
 
 	/* "serialize" all necessary information for the path private area */
 	fdw_private = serializePlanData(fdwState);
 
-	/* remove those clauses that are checked by Oracle */
-	if (fdwState->handle_clauses != NULL)
-		for (i=list_length(baserel->baserestrictinfo)-1; i>=0; --i)
-			if (fdwState->handle_clauses[i] != PUSHDOWN)
-				keep_clauses = lcons(list_nth(scan_clauses, i), keep_clauses);
+	/* keep only those clauses that are not handled by Oracle */
+	foreach(cell1, scan_clauses)
+	{
+		i = 0;
+		foreach(cell2, baserel->baserestrictinfo)
+		{
+			if (equal(lfirst(cell1), lfirst(cell2)) && fdwState->handle_clauses[i] != PUSHDOWN)
+			{
+				keep_clauses = lcons(lfirst(cell1), keep_clauses);
+				break;
+			}
+			++i;
+		}
+	}
 
 	/* remove the RestrictInfo node from all remaining clauses */
 	keep_clauses = extract_actual_clauses(keep_clauses, false);
 
 	/* Create the ForeignScan node */
-	return make_foreignscan(tlist, keep_clauses, baserel->relid, NIL, fdw_private);
+	return make_foreignscan(tlist, keep_clauses, baserel->relid, fdwState->params, fdw_private);
 }
 
 bool
@@ -565,17 +577,73 @@ oracleExplainForeignScan(ForeignScanState *node, ExplainState *es)
 void
 oracleBeginForeignScan(ForeignScanState *node, int eflags)
 {
+	ForeignScan *fsplan = (ForeignScan *)node->ss.ps.plan;
 #ifdef OLD_FDW_API
-	List *fdw_private = ((FdwPlan *)((ForeignScan *)node->ss.ps.plan)->fdwplan)->fdw_private;
+	List *fdw_private = ((FdwPlan *)fsplan->fdwplan)->fdw_private;
 #else
-	List *fdw_private = ((ForeignScan *)node->ss.ps.plan)->fdw_private;
+	List *fdw_private = fsplan->fdw_private;
+	List *exec_exprs;
+	ListCell *cell;
+	int index;
 #endif
+	struct paramDesc *paramDesc;
 	struct OracleFdwState *fdw_state;
 
 	/* deserialize private plan data */
 	fdw_state = deserializePlanData(fdw_private);
 	node->fdw_state = (void *)fdw_state;
 
+#ifndef OLD_FDW_API
+	/* create an ExprState tree for the parameter expressions */
+	exec_exprs = (List *)ExecInitExpr((Expr *)fsplan->fdw_exprs, (PlanState *)node);
+
+	/* create the list of parameters */
+	index = 0;
+	foreach(cell, exec_exprs)
+	{
+		ExprState *expr = (ExprState *)lfirst(cell);
+		char parname[10];
+
+		/* count, but skip deleted entries */
+		++index;
+		if (expr == NULL)
+			continue;
+
+		/* create a new entry in the parameter list */
+		paramDesc = (struct paramDesc *)palloc(sizeof(struct paramDesc));
+		snprintf(parname, 10, ":p%d", index);
+		paramDesc->name = pstrdup(parname);
+		paramDesc->type = exprType((Node *)(expr->expr));
+
+		if (paramDesc->type == TEXTOID || paramDesc->type == VARCHAROID
+				|| paramDesc->type == BPCHAROID || paramDesc->type == CHAROID)
+			paramDesc->type = BIND_STRING;
+		else if (paramDesc->type == DATEOID || paramDesc->type == TIMESTAMPOID
+				|| paramDesc->type == TIMESTAMPTZOID)
+			paramDesc->bindType = BIND_TIMESTAMP;
+		else
+			paramDesc->bindType = BIND_NUMBER;
+
+		paramDesc->value = NULL;
+		paramDesc->node = expr;
+		paramDesc->next = fdw_state->paramList;
+		fdw_state->paramList = paramDesc;
+	}
+#endif
+
+	/* add a fake parameter ":now" if that string appears in the query */
+	if (strstr(fdw_state->query, ":now") != NULL)
+	{
+		paramDesc = (struct paramDesc *)palloc(sizeof(struct paramDesc));
+		paramDesc->name = pstrdup(":now");
+		paramDesc->type = TIMESTAMPTZOID;
+		paramDesc->bindType = BIND_TIMESTAMP;
+		paramDesc->value = NULL;
+		paramDesc->node = NULL;
+		paramDesc->next = fdw_state->paramList;
+		fdw_state->paramList = paramDesc;
+	}
+
 	elog(DEBUG1, "oracle_fdw: begin foreign table scan on %d", RelationGetRelid(node->ss.ss_currentRelation));
 
 	/* connect to Oracle database, don't start transaction for explain only */
@@ -600,6 +668,7 @@ TupleTableSlot *
 oracleIterateForeignScan(ForeignScanState *node)
 {
 	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;
+	ExprContext *econtext = node->ss.ps.ps_ExprContext;
 	EState *execstate = node->ss.ps.state;
 	int have_result;
 	struct OracleFdwState *fdw_state = (struct OracleFdwState *)node->fdw_state;
@@ -614,7 +683,7 @@ oracleIterateForeignScan(ForeignScanState *node)
 	else
 	{
 		/* fill the parameter list with the actual values in execstate */
-		char *paramInfo = setParameters(fdw_state->paramList, execstate);
+		char *paramInfo = setParameters(fdw_state->paramList, execstate, econtext);
 
 		/* execute the Oracle statement and fetch the first row */
 		elog(DEBUG1, "oracle_fdw: execute query in foreign table scan on %d%s", RelationGetRelid(node->ss.ss_currentRelation), paramInfo);
@@ -698,6 +767,7 @@ struct OracleFdwState
 	fdwState->dbserver = NULL;
 	fdwState->user = NULL;
 	fdwState->password = NULL;
+	fdwState->params = NIL;
 	fdwState->paramList = NULL;
 	fdwState->handle_clauses = NULL;
 
@@ -854,14 +924,15 @@ getColumnData(Oid foreigntableid, struct oraTable *oraTable)
  * 		As a side effect. we also mark the used columns in oraTable.
  */
 char
-*createQuery(oracleSession *session, List *columnlist, List *conditions, struct oraTable *oraTable, struct paramDesc **paramList, handleClause **handle_clauses)
+*createQuery(oracleSession *session, RelOptInfo *foreignrel, struct oraTable *oraTable, List **params, handleClause **handle_clauses)
 {
 	ListCell *cell;
 	bool first_col = true, in_quote = false;
-	int i, clause_count = -1;
-	char *where, *wherecopy, *p, md5[33];
+	int i, clause_count = -1, index;
+	char *where, *wherecopy, *p, md5[33], parname[10];
 	StringInfoData query, result;
-	struct paramDesc *param, *prev_param = NULL;
+	List *columnlist = foreignrel->reltargetlist,
+		*conditions = foreignrel->baserestrictinfo;
 
 	/* first, find all the columns to include in the select list */
 
@@ -910,7 +981,7 @@ char
 	foreach(cell, conditions)
 	{
 		/* try to convert each condition to Oracle SQL */
-		bool can_ignore = getOracleWhereClause(session, &where, ((RestrictInfo *)lfirst(cell))->clause, oraTable, paramList);
+		bool can_ignore = getOracleWhereClause(session, foreignrel, &where, ((RestrictInfo *)lfirst(cell))->clause, oraTable, params);
 		if (where != NULL) {
 			/* append new WHERE clause to query string */
 			if (first_col)
@@ -941,31 +1012,18 @@ char
 	}
 
 	/* remove all parameters that do not actually occur in the query */
-	param = *paramList;
-	while (param != NULL)
+	index = 0;
+	foreach(cell, *params)
 	{
-		if (strstr(wherecopy, param->name) == NULL)
+		++index;
+		snprintf(parname, 10, ":p%d", index);
+		if (strstr(wherecopy, parname) == NULL)
 		{
-			/* remove parameter from linked list */
-			if (prev_param == NULL)
-				*paramList = param->next;
-			else
-				prev_param->next = param->next;
-
-			/* free memory */
-			pfree(param->name);
-			pfree(param);
+			/* set the element to NULL to indicate it's gone */
+			lfirst(cell) = NULL;
 		}
-		else
-		{
-			prev_param = param;
-		}
-
-		if (prev_param == NULL)
-			param = *paramList;
-		else
-			param = prev_param->next;
 	}
+
 	pfree(wherecopy);
 
 	/*
@@ -1130,7 +1188,7 @@ acquireSampleRowsFunc(Relation relation, int elevel, HeapTuple *rows, int targro
  * 		will be stored in paramList.
  */
 bool
-getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const struct oraTable *oraTable, struct paramDesc **paramList)
+getOracleWhereClause(oracleSession *session, RelOptInfo *foreignrel, char **where, Expr *expr, const struct oraTable *oraTable, List **params)
 {
 	char *opername, *left, *right, *arg, oprkind;
 	Const *constant;
@@ -1141,7 +1199,6 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 	CoalesceExpr *coalesceexpr;
 	CoerceViaIO *coerce;
 	Param *param;
-	struct paramDesc *paramDesc;
 	Var *variable;
 	FuncExpr *func;
 	regproc typoutput;
@@ -1191,103 +1248,132 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			break;
 		case T_Param:
 			param = (Param *)expr;
-
+#ifdef OLD_FDW_API
+			/* don't try to push down parameters with 9.1 */
+			return false;
+#else
 			/* don't try to handle interval parameters */
 			if (! canHandleType(param->paramtype) || param->paramtype == INTERVALOID)
 				return false;
 
-			/*
-			 * external parameters will be called :e1, :e2, etc.
-			 * internal parameters will be called :i0, :i1, etc.
-			 */
+			/* find the index in the parameter list */
+			index = 0;
+			foreach(cell, *params)
+			{
+				++index;
+				if (equal(param, (Node *)lfirst(cell)))
+					break;
+			}
+			if (cell == NULL)
+			{
+				/* add the parameter to the list */
+				++index;
+				*params = lappend(*params, param);
+			}
+
+			/* parameters will be called :p1, :p2 etc. */
 			initStringInfo(&result);
-			appendStringInfo(&result,
-					":%c%d",
-					(param->paramkind == PARAM_EXTERN ? 'e' : 'i'),
-					param->paramid);
-
-			/* create a new entry in the parameter list */
-			paramDesc = (struct paramDesc *)palloc(sizeof(struct paramDesc));
-			paramDesc->name = pstrdup(result.data);
-			paramDesc->number = param->paramid;
-			paramDesc->type = param->paramtype;
-
-			if (param->paramtype == TEXTOID || param->paramtype == VARCHAROID
-					|| param->paramtype == BPCHAROID || param->paramtype == CHAROID)
-				paramDesc->bindType = BIND_STRING;
-			else if (param->paramtype == DATEOID || param->paramtype == TIMESTAMPOID
-					|| param->paramtype == TIMESTAMPTZOID)
-				paramDesc->bindType = BIND_TIMESTAMP;
-			else
-				paramDesc->bindType = BIND_NUMBER;
-
-			paramDesc->value = NULL;
-			paramDesc->isExtern = (param->paramkind == PARAM_EXTERN);
-			paramDesc->next = *paramList;
-			*paramList = paramDesc;
+			appendStringInfo(&result, ":p%d", index);
 
 			/* we have to reevaluate conditions with internal parameters */
 			if (param->paramkind != PARAM_EXTERN)
 				can_ignore = false;
 
 			break;
+#endif
 		case T_Var:
 			variable = (Var *)expr;
 
-			/* we cannot handle system columns */
-			if (variable->varattno < 1)
-				return false;
-
-			/*
-			 * Allow boolean columns here.
-			 * They will be rendered as ("COL" <> 0).
-			 */
-			if (! (canHandleType(variable->vartype) || variable->vartype == BOOLOID))
-				return false;
-
-			/* get oraTable column index corresponding to this column (-1 if none) */
-			index = oraTable->ncols - 1;
-			while (index >= 0 && oraTable->cols[index]->pgattnum != variable->varattno)
-				--index;
-
-			/* if no Oracle column corresponds, translate as NULL */
-			if (index == -1)
+			if (variable->varno == foreignrel->relid && variable->varlevelsup == 0)
 			{
+				/* the variable belongs to the foreign table, replace it with the name */
+
+				/* we cannot handle system columns */
+				if (variable->varattno < 1)
+					return false;
+	
+				/*
+				 * Allow boolean columns here.
+				 * They will be rendered as ("COL" <> 0).
+				 */
+				if (! (canHandleType(variable->vartype) || variable->vartype == BOOLOID))
+					return false;
+	
+				/* get oraTable column index corresponding to this column (-1 if none) */
+				index = oraTable->ncols - 1;
+				while (index >= 0 && oraTable->cols[index]->pgattnum != variable->varattno)
+					--index;
+	
+				/* if no Oracle column corresponds, translate as NULL */
+				if (index == -1)
+				{
+					initStringInfo(&result);
+					appendStringInfo(&result, "NULL");
+					break;
+				}
+	
+				/*
+				 * Don't try to convert a column reference if the type is
+				 * converted from a non-string type in Oracle to a string type
+				 * in PostgreSQL because functions and operators won't work the same.
+				 */
+				oratype = oraTable->cols[index]->oratype;
+				if ((variable->vartype == TEXTOID
+						|| variable->vartype == BPCHAROID
+						|| variable->vartype == VARCHAROID)
+						&& oratype != ORA_TYPE_VARCHAR2
+						&& oratype != ORA_TYPE_CHAR
+						&& oratype != ORA_TYPE_NVARCHAR2
+						&& oratype != ORA_TYPE_NCHAR
+						&& oratype != ORA_TYPE_CLOB)
+					return false;
+	
 				initStringInfo(&result);
-				appendStringInfo(&result, "NULL");
-				break;
+	
+				/* work around the lack of booleans in Oracle */
+				if (variable->vartype == BOOLOID)
+				{
+					appendStringInfo(&result, "(");
+				}
+	
+				appendStringInfo(&result, "%s", oraTable->cols[index]->name);
+	
+				/* work around the lack of booleans in Oracle */
+				if (variable->vartype == BOOLOID)
+				{
+					appendStringInfo(&result, " <> 0)");
+				}
 			}
-
-			/*
-			 * Don't try to convert a column reference if the type is
-			 * converted from a non-string type in Oracle to a string type
-			 * in PostgreSQL because functions and operators won't work the same.
-			 */
-			oratype = oraTable->cols[index]->oratype;
-			if ((variable->vartype == TEXTOID
-					|| variable->vartype == BPCHAROID
-					|| variable->vartype == VARCHAROID)
-					&& oratype != ORA_TYPE_VARCHAR2
-					&& oratype != ORA_TYPE_CHAR
-					&& oratype != ORA_TYPE_NVARCHAR2
-					&& oratype != ORA_TYPE_NCHAR
-					&& oratype != ORA_TYPE_CLOB)
+			else
+			{
+				/* treat it like a parameter */
+#ifdef OLD_FDW_API
+				/* don't try to push down parameters with 9.1 */
 				return false;
-
-			initStringInfo(&result);
-
-			/* work around the lack of booleans in Oracle */
-			if (variable->vartype == BOOLOID)
-			{
-				appendStringInfo(&result, "(");
-			}
-
-			appendStringInfo(&result, "%s", oraTable->cols[index]->name);
-
-			/* work around the lack of booleans in Oracle */
-			if (variable->vartype == BOOLOID)
-			{
-				appendStringInfo(&result, " <> 0)");
+#else
+				/* don't try to handle type interval */
+				if (! canHandleType(variable->vartype) || variable->vartype == INTERVALOID)
+					return false;
+	
+				/* find the index in the parameter list */
+				index = 0;
+				foreach(cell, *params)
+				{
+					++index;
+					if (equal(variable, (Node *)lfirst(cell)))
+						break;
+				}
+				if (cell == NULL)
+				{
+					/* add the parameter to the list */
+					++index;
+					*params = lappend(*params, variable);
+				}
+	
+				/* parameters will be called :p1, :p2 etc. */
+				initStringInfo(&result);
+				appendStringInfo(&result, ":p%d", index);
+#endif
 			}
 
 			break;
@@ -1349,7 +1435,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 				|| strcmp(opername, "|/") == 0
 				|| strcmp(opername, "@") == 0)
 			{
-				can_ignore = can_ignore && getOracleWhereClause(session, &left, linitial(oper->args), oraTable, paramList);
+				can_ignore = getOracleWhereClause(session, foreignrel, &left, linitial(oper->args), oraTable, params) && can_ignore;
 				if (left == NULL)
 				{
 					pfree(opername);
@@ -1359,7 +1445,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 				if (oprkind == 'b')
 				{
 					/* binary operator */
-					can_ignore = can_ignore && getOracleWhereClause(session, &right, lsecond(oper->args), oraTable, paramList);
+					can_ignore = getOracleWhereClause(session, foreignrel, &right, lsecond(oper->args), oraTable, params) && can_ignore;
 					if (right == NULL)
 					{
 						pfree(left);
@@ -1468,15 +1554,19 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			if (! canHandleType(leftargtype))
 				return false;
 
-			can_ignore = can_ignore && getOracleWhereClause(session, &left, linitial(arrayoper->args), oraTable, paramList);
+			can_ignore = getOracleWhereClause(session, foreignrel, &left, linitial(arrayoper->args), oraTable, params) && can_ignore;
 			if (left == NULL)
 				return false;
 
+			/* only push down IN expressions with constant second (=last) argument */
+			if (((Expr *)llast(arrayoper->args))->type != T_Const)
+				return false;
+
 			/* begin to compose result */
 			initStringInfo(&result);
 			appendStringInfo(&result, "(%s %s (", left, arrayoper->useOr ? "IN" : "NOT IN");
 
-			/* the second and last argument must be a Const of ArrayType */
+			/* the second (=last) argument must be a Const of ArrayType */
 			constant = (Const *)llast(arrayoper->args);
 
 			/* loop through the array elements */
@@ -1525,12 +1615,12 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			if (! canHandleType(rightargtype))
 				return false;
 
-			can_ignore = can_ignore && getOracleWhereClause(session, &left, linitial(((DistinctExpr *)expr)->args), oraTable, paramList);
+			can_ignore = getOracleWhereClause(session, foreignrel, &left, linitial(((DistinctExpr *)expr)->args), oraTable, params) && can_ignore;
 			if (left == NULL)
 			{
 				return false;
 			}
-			can_ignore = can_ignore && getOracleWhereClause(session, &right, lsecond(((DistinctExpr *)expr)->args), oraTable, paramList);
+			can_ignore = getOracleWhereClause(session, foreignrel, &right, lsecond(((DistinctExpr *)expr)->args), oraTable, params) && can_ignore;
 			if (right == NULL)
 			{
 				pfree(left);
@@ -1554,12 +1644,12 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			if (! canHandleType(rightargtype))
 				return false;
 
-			can_ignore = can_ignore && getOracleWhereClause(session, &left, linitial(((NullIfExpr *)expr)->args), oraTable, paramList);
+			can_ignore = getOracleWhereClause(session, foreignrel, &left, linitial(((NullIfExpr *)expr)->args), oraTable, params) && can_ignore;
 			if (left == NULL)
 			{
 				return false;
 			}
-			can_ignore = can_ignore && getOracleWhereClause(session, &right, lsecond(((NullIfExpr *)expr)->args), oraTable, paramList);
+			can_ignore = getOracleWhereClause(session, foreignrel, &right, lsecond(((NullIfExpr *)expr)->args), oraTable, params) && can_ignore;
 			if (right == NULL)
 			{
 				pfree(left);
@@ -1573,7 +1663,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 		case T_BoolExpr:
 			boolexpr = (BoolExpr *)expr;
 
-			can_ignore = can_ignore && getOracleWhereClause(session, &arg, linitial(boolexpr->args), oraTable, paramList);
+			can_ignore = getOracleWhereClause(session, foreignrel, &arg, linitial(boolexpr->args), oraTable, params) && can_ignore;
 			if (arg == NULL)
 				return false;
 
@@ -1584,7 +1674,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 
 			for_each_cell(cell, lnext(list_head(boolexpr->args)))
 			{
-				can_ignore = can_ignore && getOracleWhereClause(session, &arg, (Expr *)lfirst(cell), oraTable, paramList);
+				can_ignore = getOracleWhereClause(session, foreignrel, &arg, (Expr *)lfirst(cell), oraTable, params) && can_ignore;
 				if (arg == NULL)
 				{
 					pfree(result.data);
@@ -1599,10 +1689,10 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 
 			break;
 		case T_RelabelType:
-			return getOracleWhereClause(session, where, ((RelabelType *)expr)->arg, oraTable, paramList);
+			return getOracleWhereClause(session, foreignrel, where, ((RelabelType *)expr)->arg, oraTable, params);
 			break;
 		case T_CoerceToDomain:
-			return getOracleWhereClause(session, where, ((CoerceToDomain *)expr)->arg, oraTable, paramList);
+			return getOracleWhereClause(session, foreignrel, where, ((CoerceToDomain *)expr)->arg, oraTable, params);
 			break;
 		case T_CaseExpr:
 			caseexpr = (CaseExpr *)expr;
@@ -1616,7 +1706,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			/* for the form "CASE arg WHEN ...", add first expression */
 			if (caseexpr->arg != NULL)
 			{
-				can_ignore = can_ignore && getOracleWhereClause(session, &arg, caseexpr->arg, oraTable, paramList);
+				can_ignore = getOracleWhereClause(session, foreignrel, &arg, caseexpr->arg, oraTable, params) && can_ignore;
 				if (arg == NULL)
 				{
 					pfree(result.data);
@@ -1637,12 +1727,12 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 				if (caseexpr->arg == NULL)
 				{
 					/* for CASE WHEN ..., use the whole expression */
-					can_ignore = can_ignore && getOracleWhereClause(session, &arg, whenclause->expr, oraTable, paramList);
+					can_ignore = getOracleWhereClause(session, foreignrel, &arg, whenclause->expr, oraTable, params) && can_ignore;
 				}
 				else
 				{
 					/* for CASE arg WHEN ..., use only the right branch of the equality */
-					can_ignore = can_ignore && getOracleWhereClause(session, &arg, lsecond(((OpExpr *)whenclause->expr)->args), oraTable, paramList);
+					can_ignore = getOracleWhereClause(session, foreignrel, &arg, lsecond(((OpExpr *)whenclause->expr)->args), oraTable, params) && can_ignore;
 				}
 
 				if (arg == NULL)
@@ -1657,7 +1747,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 				}
 
 				/* THEN */
-				can_ignore = can_ignore && getOracleWhereClause(session, &arg, whenclause->result, oraTable, paramList);
+				can_ignore = getOracleWhereClause(session, foreignrel, &arg, whenclause->result, oraTable, params) && can_ignore;
 				if (arg == NULL)
 				{
 					pfree(result.data);
@@ -1673,7 +1763,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			/* append ELSE clause if appropriate */
 			if (caseexpr->defresult != NULL)
 			{
-				can_ignore = can_ignore && getOracleWhereClause(session, &arg, caseexpr->defresult, oraTable, paramList);
+				can_ignore = getOracleWhereClause(session, foreignrel, &arg, caseexpr->defresult, oraTable, params) && can_ignore;
 				if (arg == NULL)
 				{
 					pfree(result.data);
@@ -1701,7 +1791,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			first_arg = true;
 			foreach(cell, coalesceexpr->args)
 			{
-				can_ignore = can_ignore && getOracleWhereClause(session, &arg, (Expr *)lfirst(cell), oraTable, paramList);
+				can_ignore = getOracleWhereClause(session, foreignrel, &arg, (Expr *)lfirst(cell), oraTable, params) && can_ignore;
 				if (arg == NULL)
 				{
 					pfree(result.data);
@@ -1724,7 +1814,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 
 			break;
 		case T_NullTest:
-			can_ignore = can_ignore && getOracleWhereClause(session, &arg, ((NullTest *)expr)->arg, oraTable, paramList);
+			can_ignore = getOracleWhereClause(session, foreignrel, &arg, ((NullTest *)expr)->arg, oraTable, params) && can_ignore;
 			if (arg == NULL)
 				return false;
 
@@ -1741,7 +1831,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 
 			/* do nothing for implicit casts */
 			if (func->funcformat == COERCE_IMPLICIT_CAST)
-				return getOracleWhereClause(session, where, linitial(func->args), oraTable, paramList);
+				return getOracleWhereClause(session, foreignrel, where, linitial(func->args), oraTable, params);
 
 			/* get function name and schema */
 			tuple = SearchSysCache1(PROCOID, ObjectIdGetDatum(func->funcid));
@@ -1821,7 +1911,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 				first_arg = true;
 				foreach(cell, func->args)
 				{
-					can_ignore = can_ignore && getOracleWhereClause(session, &arg, lfirst(cell), oraTable, paramList);
+					can_ignore = getOracleWhereClause(session, foreignrel, &arg, lfirst(cell), oraTable, params) && can_ignore;
 					if (arg == NULL)
 					{
 						pfree(result.data);
@@ -1846,7 +1936,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 			else if (strcmp(opername, "date_part") == 0)
 			{
 				/* special case: EXTRACT */
-				can_ignore = can_ignore && getOracleWhereClause(session, &left, linitial(func->args), oraTable, paramList);
+				can_ignore = getOracleWhereClause(session, foreignrel, &left, linitial(func->args), oraTable, params) && can_ignore;
 				if (left == NULL)
 				{
 					pfree(opername);
@@ -1866,7 +1956,7 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 					/* remove final quote */
 					left[strlen(left) - 1] = '\0';
 
-					can_ignore = can_ignore && getOracleWhereClause(session, &right, lsecond(func->args), oraTable, paramList);
+					can_ignore = getOracleWhereClause(session, foreignrel, &right, lsecond(func->args), oraTable, params) && can_ignore;
 					if (right == NULL)
 					{
 						pfree(opername);
@@ -1892,17 +1982,6 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 				/* special case: current timestamp */
 				initStringInfo(&result);
 				appendStringInfo(&result, ":now");
-
-				/* create a new entry in the parameter list */
-				paramDesc = (struct paramDesc *)palloc(sizeof(struct paramDesc));
-				paramDesc->name = pstrdup(":now");
-				paramDesc->number = 0;
-				paramDesc->type = TIMESTAMPTZOID;
-				paramDesc->bindType = BIND_TIMESTAMP;
-				paramDesc->value = NULL;
-				paramDesc->isExtern = 1;
-				paramDesc->next = *paramList;
-				*paramList = paramDesc;
 			}
 			else
 			{
@@ -1931,27 +2010,34 @@ getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const str
 
 			/* the argument must be a not-NULL text constant */
 			constant = (Const *)coerce->arg;
-			if (constant->constisnull || constant->consttype != TEXTOID)
+			if (constant->constisnull || (constant->consttype != CSTRINGOID && constant->consttype != TEXTOID))
 				return false;
 
+			/* get the type's output function */
+			tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(constant->consttype));
+			if (!HeapTupleIsValid(tuple))
+			{
+				elog(ERROR, "cache lookup failed for type %u", constant->consttype);
+			}
+			typoutput = ((Form_pg_type)GETSTRUCT(tuple))->typoutput;
+			ReleaseSysCache(tuple);
+
 			/* the value must be "now" */
-			if (VARSIZE(constant->constvalue) - VARHDRSZ != 3
-					|| strncmp(VARDATA(constant->constvalue), "now", 3) != 0)
+			if (strcmp(DatumGetCString(OidFunctionCall1(typoutput, constant->constvalue)), "now") != 0)
 				return false;
 
 			initStringInfo(&result);
-			appendStringInfo(&result, ":now");
-
-			/* create a new entry in the parameter list */
-			paramDesc = (struct paramDesc *)palloc(sizeof(struct paramDesc));
-			paramDesc->name = pstrdup(":now");
-			paramDesc->number = 0;
-			paramDesc->type = coerce->resulttype;
-			paramDesc->bindType = BIND_TIMESTAMP;
-			paramDesc->value = NULL;
-			paramDesc->isExtern = 1;
-			paramDesc->next = *paramList;
-			*paramList = paramDesc;
+			switch (coerce->resulttype)
+			{
+				case DATEOID:
+					appendStringInfo(&result, "(CAST (:now AS DATE))");
+					break;
+				case TIMESTAMPOID:
+					appendStringInfo(&result, "(CAST (:now AS TIMESTAMP))");
+					break;
+				case TIMESTAMPTZOID:
+					appendStringInfo(&result, ":now");
+			}
 
 			break;
 		default:
@@ -2515,8 +2601,7 @@ List
 *serializePlanData(struct OracleFdwState *fdwState)
 {
 	List *result = NIL;
-	int i, len = 0;
-	const struct paramDesc *param;
+	int i;
 
 	/* dbserver */
 	result = lappend(result, serializeString(fdwState->dbserver));
@@ -2550,22 +2635,7 @@ List
 		result = lappend(result, serializeLong(fdwState->oraTable->cols[i]->val_size));
 		/* don't serialize val, val_len and val_null */
 	}
-	/* find length of parameter list */
-	for (param=fdwState->paramList; param; param=param->next)
-		++len;
-	/* serialize length */
-	result = lappend(result, serializeInt(len));
-	/* parameter list entries */
-	for (param=fdwState->paramList; param; param=param->next)
-	{
-		result = lappend(result, serializeString(param->name));
-		result = lappend(result, serializeInt(param->number));
-		result = lappend(result, serializeOid(param->type));
-		result = lappend(result, serializeInt((int)param->bindType));
-		/* don't serialize value */
-		result = lappend(result, serializeInt(param->isExtern));
-	}
-	/* don't serialize startup_cost, total_cost, handle_clauses, rowcount and columnindex */
+	/* don't serialize params, paramList, startup_cost, total_cost, handle_clauses, rowcount and columnindex */
 
 	return result;
 }
@@ -2622,8 +2692,7 @@ struct OracleFdwState
 {
 	struct OracleFdwState *state = palloc(sizeof(struct OracleFdwState));
 	ListCell *cell = list_head(list);
-	int i, len;
-	struct paramDesc *param;
+	int i;
 
 	/* session will be set upon connect */
 	state->session = NULL;
@@ -2634,6 +2703,8 @@ struct OracleFdwState
 	/* these are not serialized */
 	state->rowcount = 0;
 	state->columnindex = 0;
+	state->params = NULL;
+	state->paramList = NULL;
 
 	/* dbserver */
 	state->dbserver = deserializeString(lfirst(cell));
@@ -2695,30 +2766,6 @@ struct OracleFdwState
 		state->oraTable->cols[i]->val_null = 1;
 	}
 
-	/* length of parameter list */
-	len = (int)DatumGetInt32(((Const *)lfirst(cell))->constvalue);
-	cell = lnext(cell);
-
-	/* parameter table entries */
-	state->paramList = NULL;
-	for (i=0; i<len; ++i)
-	{
-		param = (struct paramDesc *)palloc(sizeof(struct paramDesc));
-		param->name = deserializeString(lfirst(cell));
-		cell = lnext(cell);
-		param->number = (int)DatumGetInt32(((Const *)lfirst(cell))->constvalue);
-		cell = lnext(cell);
-		param->type = DatumGetObjectId(((Const *)lfirst(cell))->constvalue);
-		cell = lnext(cell);
-		param->bindType = (oraBindType)DatumGetInt32(((Const *)lfirst(cell))->constvalue);
-		cell = lnext(cell);
-		param->value = NULL;
-		param->isExtern = (int)DatumGetInt32(((Const *)lfirst(cell))->constvalue);
-		cell = lnext(cell);
-		param->next = state->paramList;
-		state->paramList = param;
-	}
-
 	return state;
 }
 
@@ -2786,67 +2833,43 @@ exitHook(int code, Datum arg)
  * 		Return a string containing the parameters set for a DEBUG message.
  */
 char *
-setParameters(struct paramDesc *paramList, EState *execstate)
+setParameters(struct paramDesc *paramList, EState *execstate, ExprContext *econtext)
 {
 	struct paramDesc *param;
 	Datum datum;
 	HeapTuple tuple;
 	TimestampTz tstamp;
-	int is_null;
+	bool is_null;
 	bool first_param = true;
+#ifndef OLD_FDW_API
+	MemoryContext oldcontext;
+#endif
 	StringInfoData info;  /* list of parameters for DEBUG message */
 	initStringInfo(&info);
 
-	/*
- 	* There is no open statement yet.
- 	* Run the query and get the first result row.
- 	*/
+#ifndef OLD_FDW_API
+	/* switch to short lived memory context */
+	oldcontext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);
+#endif
 
 	/* iterate parameter list and fill values */
 	for (param=paramList; param; param=param->next)
 	{
 		if (strcmp(param->name, ":now") == 0)
 		{
-			/*
-		 	* This parameter will be set to the transaction start timestamp.
-		 	*/
-			regproc castfunc;
-
 			/* get transaction start timestamp */
 			tstamp = GetCurrentTransactionStartTimestamp();
 
-			if (param->type == TIMESTAMPTZOID)
-			{
-				/* no need to cast */
-				datum = TimestampGetDatum(tstamp);
-			}
-			else
-			{
-				/* find the cast function to the desired target type */
-				tuple = SearchSysCache2(CASTSOURCETARGET, ObjectIdGetDatum(TIMESTAMPTZOID), ObjectIdGetDatum(param->type));
-				if (!HeapTupleIsValid(tuple))
-				{
-					elog(ERROR, "cache lookup failed for cast from %u to %u", TIMESTAMPTZOID, param->type);
-				}
-				castfunc = ((Form_pg_cast)GETSTRUCT(tuple))->castfunc;
-				ReleaseSysCache(tuple);
-
-				/* cast */
-				datum = OidFunctionCall1(castfunc, TimestampGetDatum(tstamp));
-			}
-			is_null = 0;
-		}
-		else if (param->isExtern)
-		{
-			/* external parameters are numbered from 1 on */
-			datum = execstate->es_param_list_info->params[param->number-1].value;
-			is_null = execstate->es_param_list_info->params[param->number-1].isnull;
+			datum = TimestampGetDatum(tstamp);
+			is_null = false;
 		}
 		else
 		{
-			/* internal parameters are numbered from 0 on */
-			datum = execstate->es_param_exec_vals[param->number].value;
-			is_null = execstate->es_param_exec_vals[param->number].isnull;
+			/*
+			 * Evaluate the expression.
+			 * This code path cannot be reached in 9.1
+			 */
+			datum = ExecEvalExpr((ExprState *)(param->node), econtext, &is_null, NULL);
 		}
 
 		if (is_null)
@@ -2882,6 +2905,11 @@ setParameters(struct paramDesc *paramList, EState *execstate)
 		}
 	}
 
+#ifndef OLD_FDW_API
+	/* reset memory context */
+	MemoryContextSwitchTo(oldcontext);
+#endif
+
 	return info.data;
 }
 
@@ -2897,12 +2925,12 @@ convertTuple(struct OracleFdwState *fdw_state, Datum *values, bool *nulls, bool
 	char *value = NULL;
 	long value_len = 0;
 	int j, index = -1;
-	ErrorContextCallback errcontext;
+	ErrorContextCallback errcb;
 	Oid pgtype;
 
 	/* initialize error context callback, install it only during conversions */
-	errcontext.callback = errorContextCallback;
-	errcontext.arg = (void *)fdw_state;
+	errcb.callback = errorContextCallback;
+	errcb.arg = (void *)fdw_state;
 
 	/* assign result values */
 	for (j=0; j<fdw_state->oraTable->npgcols; ++j)
@@ -2989,8 +3017,8 @@ convertTuple(struct OracleFdwState *fdw_state, Datum *values, bool *nulls, bool
 			dat = CStringGetDatum(value);
 
 			/* install error context callback */
-			errcontext.previous = error_context_stack;
-			error_context_stack = &errcontext;
+			errcb.previous = error_context_stack;
+			error_context_stack = &errcb;
 			fdw_state->columnindex = index;
 
 			/* for string types, check that the data are in the database encoding */
@@ -3018,7 +3046,7 @@ convertTuple(struct OracleFdwState *fdw_state, Datum *values, bool *nulls, bool
 			}
 
 			/* uninstall error context callback */
-			error_context_stack = errcontext.previous;
+			error_context_stack = errcb.previous;
 		}
 
 		/* free the data buffer for LOBs */
